package org.apache.bcel.util;

package .tmp;

import java.io.Closeable;
import java.io.DataInputStream;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Locale;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.zip.ZipFile;
import org.apache.bcel.util.ModularRuntimeImage;

public class ClassPath implements Closeable {
  private static final FilenameFilter ARCHIVE_FILTER;
  
  private static final FilenameFilter MODULES_FILTER;
  
  static {
    ARCHIVE_FILTER = ((dir, name) -> {
        name = name.toLowerCase(Locale.ENGLISH);
        return (name.endsWith(".zip") || name.endsWith(".jar"));
      });
    MODULES_FILTER = ((dir, name) -> {
        name = name.toLowerCase(Locale.ENGLISH);
        return name.endsWith(".jmod");
      });
  }
  
  public static final org.apache.bcel.util.ClassPath SYSTEM_CLASS_PATH = new org.apache.bcel.util.ClassPath(getClassPath());
  
  private final String classPath;
  
  private org.apache.bcel.util.ClassPath parent;
  
  private final AbstractPathEntry[] paths;
  
  private static void addJdkModules(String javaHome, List<String> list) {
    String modulesPath = System.getProperty("java.modules.path");
    if (modulesPath == null || modulesPath.trim().isEmpty())
      modulesPath = javaHome + File.separator + "jmods"; 
    File modulesDir = new File(modulesPath);
    if (modulesDir.exists()) {
      String[] modules = modulesDir.list(MODULES_FILTER);
      if (modules != null)
        for (String module : modules)
          list.add(modulesDir.getPath() + File.separatorChar + module);  
    } 
  }
  
  public static String getClassPath() {
    String classPathProp = System.getProperty("java.class.path");
    String bootClassPathProp = System.getProperty("sun.boot.class.path");
    String extDirs = System.getProperty("java.ext.dirs");
    String javaHome = System.getProperty("java.home");
    List<String> list = new ArrayList<>();
    Path modulesPath = Paths.get(javaHome, new String[0]).resolve("lib/modules");
    if (Files.exists(modulesPath, new java.nio.file.LinkOption[0]) && Files.isRegularFile(modulesPath, new java.nio.file.LinkOption[0]))
      list.add(modulesPath.toAbsolutePath().toString()); 
    addJdkModules(javaHome, list);
    getPathComponents(classPathProp, list);
    getPathComponents(bootClassPathProp, list);
    List<String> dirs = new ArrayList<>();
    getPathComponents(extDirs, dirs);
    for (String d : dirs) {
      File extDir = new File(d);
      String[] extensions = extDir.list(ARCHIVE_FILTER);
      if (extensions != null)
        for (String extension : extensions)
          list.add(extDir.getPath() + File.separatorChar + extension);  
    } 
    StringBuilder buf = new StringBuilder();
    String separator = "";
    for (String path : list) {
      buf.append(separator);
      separator = File.pathSeparator;
      buf.append(path);
    } 
    return buf.toString().intern();
  }
  
  private static void getPathComponents(String path, List<String> list) {
    if (path != null) {
      StringTokenizer tokenizer = new StringTokenizer(path, File.pathSeparator);
      while (tokenizer.hasMoreTokens()) {
        String name = tokenizer.nextToken();
        File file = new File(name);
        if (file.exists())
          list.add(name); 
      } 
    } 
  }
  
  static String packageToFolder(String name) {
    return name.replace('.', '/');
  }
  
  @Deprecated
  public ClassPath() {
    this(getClassPath());
  }
  
  public ClassPath(org.apache.bcel.util.ClassPath parent, String classPath) {
    this(classPath);
    this.parent = parent;
  }
  
  public ClassPath(String classPath) {
    this.classPath = classPath;
    List<AbstractPathEntry> list = new ArrayList<>();
    for (StringTokenizer tokenizer = new StringTokenizer(classPath, File.pathSeparator); tokenizer.hasMoreTokens(); ) {
      String path = tokenizer.nextToken();
      if (!path.isEmpty()) {
        File file = new File(path);
        try {
          if (file.exists()) {
            if (file.isDirectory()) {
              list.add(new Dir(path));
              continue;
            } 
            if (path.endsWith(".jmod")) {
              list.add(new Module(new ZipFile(file)));
              continue;
            } 
            if (path.endsWith(ModularRuntimeImage.MODULES_PATH)) {
              list.add(new JrtModules(ModularRuntimeImage.MODULES_PATH));
              continue;
            } 
            list.add(new Jar(new ZipFile(file)));
          } 
        } catch (IOException e) {
          if (path.endsWith(".zip") || path.endsWith(".jar"))
            System.err.println("CLASSPATH component " + file + ": " + e); 
        } 
      } 
    } 
    this.paths = new AbstractPathEntry[list.size()];
    list.toArray(this.paths);
  }
  
  public void close() throws IOException {
    if (this.paths != null)
      for (AbstractPathEntry path : this.paths)
        path.close();  
  }
  
  public boolean equals(Object o) {
    if (o instanceof org.apache.bcel.util.ClassPath) {
      org.apache.bcel.util.ClassPath cp = (org.apache.bcel.util.ClassPath)o;
      return this.classPath.equals(cp.toString());
    } 
    return false;
  }
  
  public byte[] getBytes(String name) throws IOException {
    return getBytes(name, ".class");
  }
  
  public byte[] getBytes(String name, String suffix) throws IOException {
    DataInputStream dis = null;
    try (InputStream inputStream = getInputStream(name, suffix)) {
      if (inputStream == null)
        throw new IOException("Couldn't find: " + name + suffix); 
      dis = new DataInputStream(inputStream);
      byte[] bytes = new byte[inputStream.available()];
      dis.readFully(bytes);
      return bytes;
    } finally {
      if (dis != null)
        dis.close(); 
    } 
  }
  
  public ClassFile getClassFile(String name) throws IOException {
    return getClassFile(name, ".class");
  }
  
  public ClassFile getClassFile(String name, String suffix) throws IOException {
    ClassFile cf = null;
    if (this.parent != null)
      cf = this.parent.getClassFileInternal(name, suffix); 
    if (cf == null)
      cf = getClassFileInternal(name, suffix); 
    if (cf != null)
      return cf; 
    throw new IOException("Couldn't find: " + name + suffix);
  }
  
  private ClassFile getClassFileInternal(String name, String suffix) {
    for (AbstractPathEntry path : this.paths) {
      ClassFile cf = path.getClassFile(name, suffix);
      if (cf != null)
        return cf; 
    } 
    return null;
  }
  
  public InputStream getInputStream(String name) throws IOException {
    return getInputStream(packageToFolder(name), ".class");
  }
  
  public InputStream getInputStream(String name, String suffix) throws IOException {
    InputStream inputStream = null;
    try {
      inputStream = getClass().getClassLoader().getResourceAsStream(name + suffix);
    } catch (Exception exception) {}
    if (inputStream != null)
      return inputStream; 
    return getClassFile(name, suffix).getInputStream();
  }
  
  public String getPath(String name) throws IOException {
    int index = name.lastIndexOf('.');
    String suffix = "";
    if (index > 0) {
      suffix = name.substring(index);
      name = name.substring(0, index);
    } 
    return getPath(name, suffix);
  }
  
  public String getPath(String name, String suffix) throws IOException {
    return getClassFile(name, suffix).getPath();
  }
  
  public URL getResource(String name) {
    for (AbstractPathEntry path : this.paths) {
      URL url;
      if ((url = path.getResource(name)) != null)
        return url; 
    } 
    return null;
  }
  
  public InputStream getResourceAsStream(String name) {
    for (AbstractPathEntry path : this.paths) {
      InputStream is;
      if ((is = path.getResourceAsStream(name)) != null)
        return is; 
    } 
    return null;
  }
  
  public Enumeration<URL> getResources(String name) {
    Vector<URL> results = new Vector<>();
    for (AbstractPathEntry path : this.paths) {
      URL url;
      if ((url = path.getResource(name)) != null)
        results.add(url); 
    } 
    return results.elements();
  }
  
  public int hashCode() {
    if (this.parent != null)
      return this.classPath.hashCode() + this.parent.hashCode(); 
    return this.classPath.hashCode();
  }
  
  public String toString() {
    if (this.parent != null)
      return this.parent + File.pathSeparator + this.classPath; 
    return this.classPath;
  }
}
